@@ -24,6 +24,29 @@ namespace qc {
         return nops;
     }
 
+    std::size_t QuantumComputation::getNsingleQubitOps() const {
+        std::size_t nops = 0;
+        for (const auto& op: ops) {
+            if (!op->isUnitary()) {
+                continue;
+            }
+
+            if (op->isCompoundOperation()) {
+                const auto* const comp = dynamic_cast<const CompoundOperation*>(op.get());
+                for (const auto& subop: *comp) {
+                    if (subop->isUnitary() && !subop->isControlled() && subop->getNtargets() == 1U) {
+                        ++nops;
+                    }
+                }
+            } else {
+                if (!op->isControlled() && op->getNtargets() == 1U) {
+                    ++nops;
+                }
+            }
+        }
+        return nops;
+    }
+
     void QuantumComputation::import(const std::string& filename) {
         size_t      dot       = filename.find_last_of('.');
         std::string extension = filename.substr(dot + 1);
@@ -104,20 +127,21 @@ namespace qc {
 
         for (auto opIt = ops.begin(); opIt != ops.end(); ++opIt) {
             if ((*opIt)->getType() == qc::Measure) {
-                if (!isLastOperationOnQubit(opIt)) {
-                    continue;
-                }
-
                 outputPermutationFromMeasurements = true;
                 auto op                           = dynamic_cast<NonUnitaryOperation*>(opIt->get());
                 assert(op->getTargets().size() == op->getClassics().size());
                 auto classicIt = op->getClassics().cbegin();
                 for (const auto& q: op->getTargets()) {
-                    auto qubitidx = q;
-                    auto bitidx   = *classicIt;
+                    const auto qubitidx = q;
+                    // only the first measurement of a qubit is used to determine the output permutation
+                    if (measuredQubits.count(qubitidx) != 0) {
+                        continue;
+                    }
+
+                    const auto bitidx = *classicIt;
                     if (outputPermutationFound) {
                         // output permutation was already set before -> permute existing values
-                        auto current = outputPermutation.at(qubitidx);
+                        const auto current = outputPermutation.at(qubitidx);
                         if (static_cast<std::size_t>(qubitidx) != bitidx && static_cast<std::size_t>(current) != bitidx) {
                             for (auto& p: outputPermutation) {
                                 if (static_cast<std::size_t>(p.second) == bitidx) {
@@ -177,11 +201,11 @@ namespace qc {
     }
 
     void QuantumComputation::addQubitRegister(std::size_t nq, const char* reg_name) {
-        if (static_cast<std::size_t>(nqubits + nancillae + nq) > dd::Package<>::maxPossibleQubits) {
+        if (static_cast<std::size_t>(nqubits + nancillae + nq) > dd::Package<>::MAX_POSSIBLE_QUBITS) {
             throw QFRException("Requested too many qubits to be handled by the DD package. Qubit datatype only allows up to " +
-                               std::to_string(dd::Package<>::maxPossibleQubits) + " qubits, while " +
+                               std::to_string(dd::Package<>::MAX_POSSIBLE_QUBITS) + " qubits, while " +
                                std::to_string(nqubits + nancillae + nq) + " were requested. If you want to use more than " +
-                               std::to_string(dd::Package<>::maxPossibleQubits) + " qubits, you have to recompile the package with a wider Qubit type in `export/dd_package/include/dd/Definitions.hpp!`");
+                               std::to_string(dd::Package<>::MAX_POSSIBLE_QUBITS) + " qubits, you have to recompile the package with a wider Qubit type in `export/dd_package/include/dd/Definitions.hpp!`");
         }
 
         if (qregs.count(reg_name)) {
@@ -221,11 +245,11 @@ namespace qc {
     }
 
     void QuantumComputation::addAncillaryRegister(std::size_t nq, const char* reg_name) {
-        if (static_cast<std::size_t>(nqubits + nancillae + nq) > dd::Package<>::maxPossibleQubits) {
+        if (static_cast<std::size_t>(nqubits + nancillae + nq) > dd::Package<>::MAX_POSSIBLE_QUBITS) {
             throw QFRException("Requested too many qubits to be handled by the DD package. Qubit datatype only allows up to " +
-                               std::to_string(dd::Package<>::maxPossibleQubits) + " qubits, while " +
+                               std::to_string(dd::Package<>::MAX_POSSIBLE_QUBITS) + " qubits, while " +
                                std::to_string(nqubits + nancillae + nq) + " were requested. If you want to use more than " +
-                               std::to_string(dd::Package<>::maxPossibleQubits) + " qubits, you have to recompile the package with a wider Qubit type in `export/dd_package/include/dd/Definitions.hpp!`");
+                               std::to_string(dd::Package<>::MAX_POSSIBLE_QUBITS) + " qubits, you have to recompile the package with a wider Qubit type in `export/dd_package/include/dd/Definitions.hpp!`");
         }
 
         dd::QubitCount totalqubits = nqubits + nancillae;
@@ -398,6 +422,10 @@ namespace qc {
         // index of logical qubit
         auto logical_qubit_index = static_cast<dd::Qubit>(nqubits + nancillae);
 
+        // resize ancillary and garbage tracking vectors
+        ancillary.resize(logical_qubit_index + 1U);
+        garbage.resize(logical_qubit_index + 1U);
+
         // increase ancillae count and mark as ancillary
         nancillae++;
         ancillary[logical_qubit_index] = true;
@@ -408,6 +436,9 @@ namespace qc {
         // adjust output permutation
         if (output_qubit_index >= 0) {
             outputPermutation.insert({physical_qubit_index, output_qubit_index});
+        } else {
+            // if a qubit is not relevant for the output, it is considered garbage
+            garbage[logical_qubit_index] = true;
         }
 
         // update all operations
@@ -944,7 +975,20 @@ namespace qc {
         }
     }
 
-    bool QuantumComputation::isLastOperationOnQubit(const decltype(ops.cbegin())& opIt, const decltype(ops.cend())& end) const {
+    [[nodiscard]] std::pair<bool, std::optional<dd::Qubit>> QuantumComputation::containsLogicalQubit(const dd::Qubit logicalQubitIndex) const {
+        if (const auto it = std::find_if(
+                    initialLayout.cbegin(),
+                    initialLayout.cend(),
+                    [&logicalQubitIndex](const auto& mapping) {
+                        return mapping.second == logicalQubitIndex;
+                    });
+            it != initialLayout.cend()) {
+            return {true, it->first};
+        }
+        return {false, {}};
+    }
+
+    bool QuantumComputation::isLastOperationOnQubit(const const_iterator& opIt, const const_iterator& end) const {
         if (opIt == end)
             return true;
 
@@ -966,4 +1010,178 @@ namespace qc {
         }
         return true;
     }
+
+    void QuantumComputation::unifyQuantumRegisters(const std::string& regName) {
+        ancregs.clear();
+        qregs.clear();
+        qregs[regName] = {0, getNqubits()};
+        nancillae      = 0;
+    }
+
+    void QuantumComputation::appendMeasurementsAccordingToOutputPermutation(const std::string& registerName) {
+        // ensure that the circuit contains enough classical registers
+        if (cregs.empty()) {
+            // in case there are no registers, create a new one
+            addClassicalRegister(outputPermutation.size(), registerName.c_str());
+        } else if (nclassics < outputPermutation.size()) {
+            if (cregs.find(registerName) == cregs.end()) {
+                // in case there are registers but not enough, add a new one
+                addClassicalRegister(outputPermutation.size() - nclassics, registerName.c_str());
+            } else {
+                // in case the register already exists, augment it
+                nclassics += outputPermutation.size() - nclassics;
+                cregs[registerName].second = outputPermutation.size();
+            }
+        }
+
+        // append measurements according to output permutation
+        for (const auto& [qubit, clbit]: outputPermutation) {
+            measure(qubit, clbit);
+        }
+    }
+
+    void QuantumComputation::checkQubitRange(dd::Qubit qubit) const {
+        if (const auto it = initialLayout.find(qubit); it == initialLayout.end() || it->second >= getNqubits())
+            throw QFRException("Qubit index out of range: " +
+                               std::to_string(qubit));
+    }
+    void QuantumComputation::checkQubitRange(dd::Qubit qubit0, dd::Qubit qubit1) const {
+        checkQubitRange(qubit0);
+        checkQubitRange(qubit1);
+    }
+    void QuantumComputation::checkQubitRange(dd::Qubit qubit, const dd::Control& control) const {
+        checkQubitRange(qubit);
+        checkQubitRange(control.qubit);
+    }
+    void QuantumComputation::checkQubitRange(dd::Qubit qubit0, dd::Qubit qubit1, const dd::Control& control) const {
+        checkQubitRange(qubit0, qubit1);
+        checkQubitRange(control.qubit);
+    }
+    void QuantumComputation::checkQubitRange(dd::Qubit qubit, const dd::Controls& controls) const {
+        checkQubitRange(qubit);
+        for (auto& [ctrl, _]: controls)
+            checkQubitRange(ctrl);
+    }
+
+    void QuantumComputation::checkQubitRange(dd::Qubit qubit0, dd::Qubit qubit1, const dd::Controls& controls) const {
+        checkQubitRange(qubit0, controls);
+        checkQubitRange(qubit1);
+    }
+
+    void QuantumComputation::checkQubitRange(const std::vector<dd::Qubit>& qubits) const {
+        std::for_each(qubits.begin(), qubits.end(), [&](auto q) { checkQubitRange(q); });
+    }
+
+    void QuantumComputation::addVariable(const SymbolOrNumber& expr) {
+        if (std::holds_alternative<Symbolic>(expr)) {
+            const auto& sym = std::get<Symbolic>(expr);
+            for (const auto& term: sym)
+                occuringVariables.insert(term.getVar());
+        }
+    }
+
+    void QuantumComputation::u3(dd::Qubit target, const SymbolOrNumber& lambda, const SymbolOrNumber& phi, const SymbolOrNumber& theta) {
+        checkQubitRange(target);
+        addVariables(lambda, phi, theta);
+        emplace_back<SymbolicOperation>(getNqubits(), target, qc::U3, lambda, phi, theta);
+    }
+    void QuantumComputation::u3(dd::Qubit target, const dd::Control& control, const SymbolOrNumber& lambda, const SymbolOrNumber& phi, const SymbolOrNumber& theta) {
+        checkQubitRange(target, control);
+        addVariables(lambda, phi, theta);
+        emplace_back<SymbolicOperation>(getNqubits(), control, target, qc::U3, lambda, phi, theta);
+    }
+    void QuantumComputation::u3(dd::Qubit target, const dd::Controls& controls, const SymbolOrNumber& lambda, const SymbolOrNumber& phi, const SymbolOrNumber& theta) {
+        checkQubitRange(target, controls);
+        addVariables(lambda, phi, theta);
+        emplace_back<SymbolicOperation>(getNqubits(), controls, target, qc::U3, lambda, phi, theta);
+    }
+
+    void QuantumComputation::u2(dd::Qubit target, const SymbolOrNumber& lambda, const SymbolOrNumber& phi) {
+        checkQubitRange(target);
+        addVariables(lambda, phi);
+        emplace_back<SymbolicOperation>(getNqubits(), target, qc::U2, lambda, phi);
+    }
+    void QuantumComputation::u2(dd::Qubit target, const dd::Control& control, const SymbolOrNumber& lambda, const SymbolOrNumber& phi) {
+        checkQubitRange(target, control);
+        addVariables(lambda, phi);
+        emplace_back<SymbolicOperation>(getNqubits(), control, target, qc::U2, lambda, phi);
+    }
+    void QuantumComputation::u2(dd::Qubit target, const dd::Controls& controls, const SymbolOrNumber& lambda, const SymbolOrNumber& phi) {
+        checkQubitRange(target, controls);
+        addVariables(lambda, phi);
+        emplace_back<SymbolicOperation>(getNqubits(), controls, target, qc::U2, lambda, phi);
+    }
+
+    void QuantumComputation::phase(dd::Qubit target, const SymbolOrNumber& lambda) {
+        checkQubitRange(target);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), target, qc::Phase, lambda);
+    }
+    void QuantumComputation::phase(dd::Qubit target, const dd::Control& control, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, control);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), control, target, qc::Phase, lambda);
+    }
+    void QuantumComputation::phase(dd::Qubit target, const dd::Controls& controls, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, controls);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), controls, target, qc::Phase, lambda);
+    }
+
+    void QuantumComputation::rx(dd::Qubit target, const SymbolOrNumber& lambda) {
+        checkQubitRange(target);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), target, qc::RX, lambda);
+    }
+    void QuantumComputation::rx(dd::Qubit target, const dd::Control& control, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, control);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), control, target, qc::RX, lambda);
+    }
+    void QuantumComputation::rx(dd::Qubit target, const dd::Controls& controls, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, controls);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), controls, target, qc::RX, lambda);
+    }
+
+    void QuantumComputation::ry(dd::Qubit target, const SymbolOrNumber& lambda) {
+        checkQubitRange(target);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), target, qc::RY, lambda);
+    }
+    void QuantumComputation::ry(dd::Qubit target, const dd::Control& control, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, control);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), control, target, qc::RY, lambda);
+    }
+    void QuantumComputation::ry(dd::Qubit target, const dd::Controls& controls, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, controls);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), controls, target, qc::RY, lambda);
+    }
+
+    void QuantumComputation::rz(dd::Qubit target, const SymbolOrNumber& lambda) {
+        checkQubitRange(target);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), target, qc::RZ, lambda);
+    }
+    void QuantumComputation::rz(dd::Qubit target, const dd::Control& control, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, control);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), control, target, qc::RZ, lambda);
+    }
+    void QuantumComputation::rz(dd::Qubit target, const dd::Controls& controls, const SymbolOrNumber& lambda) {
+        checkQubitRange(target, controls);
+        addVariables(lambda);
+        emplace_back<SymbolicOperation>(getNqubits(), controls, target, qc::RZ, lambda);
+    }
+
+    // Instantiates this computation
+    void QuantumComputation::instantiate(const VariableAssignment& assignment) {
+        for (auto& op: ops) {
+            if (auto* symOp = dynamic_cast<SymbolicOperation*>(op.get())) {
+                symOp->instantiate(assignment);
+            }
+        }
+    }
 } // namespace qc
