@@ -523,6 +523,23 @@ namespace qasm {
                 return std::make_unique<qc::CompoundOperation>(std::move(gate));
             }
 
+        } else if (sym == Token::Kind::sxgate || sym == Token::Kind::sxdggate) {
+            const auto type = (sym == Token::Kind::sxgate) ? qc::SX : qc::SXdag;
+            scan();
+
+            auto target = ArgumentQreg();
+            check(Token::Kind::semicolon);
+
+            if (target.second == 1) {
+                return std::make_unique<qc::StandardOperation>(nqubits, target.first, type);
+            }
+
+            // TODO: multiple targets could be useful here
+            auto gate = qc::CompoundOperation(nqubits);
+            for (dd::QubitCount i = 0; i < target.second; ++i) {
+                gate.emplace_back<qc::StandardOperation>(nqubits, static_cast<dd::Qubit>(target.first + i), type);
+            }
+            return std::make_unique<qc::CompoundOperation>(std::move(gate));
         } else if (sym == Token::Kind::identifier) {
             scan();
             auto           gateName  = t.str;
@@ -672,7 +689,7 @@ namespace qasm {
                         for (size_t j = 0; j < parameters.size(); ++j)
                             paramMap[cGateIt->second.parameterNames[j]] = parameters[j];
 
-                        if (auto cu = dynamic_cast<Ugate*>(cGate)) {
+                        if (auto cu = dynamic_cast<SingleQubitGate*>(cGate)) {
                             std::unique_ptr<Expr> theta(RewriteExpr(cu->theta, paramMap));
                             std::unique_ptr<Expr> phi(RewriteExpr(cu->phi, paramMap));
                             std::unique_ptr<Expr> lambda(RewriteExpr(cu->lambda, paramMap));
@@ -689,7 +706,7 @@ namespace qasm {
                 // identifier specifies just a single operation (U3 or CX)
                 if (gateIt != compoundGates.end() && gateIt->second.gates.size() == 1) {
                     auto gate = gateIt->second.gates.front();
-                    if (auto u = dynamic_cast<Ugate*>(gate)) {
+                    if (auto u = dynamic_cast<SingleQubitGate*>(gate)) {
                         std::unique_ptr<Expr> theta(RewriteExpr(u->theta, paramMap));
                         std::unique_ptr<Expr> phi(RewriteExpr(u->phi, paramMap));
                         std::unique_ptr<Expr> lambda(RewriteExpr(u->lambda, paramMap));
@@ -706,13 +723,16 @@ namespace qasm {
 
                 qc::CompoundOperation op(nqubits);
                 for (auto& gate: gateIt->second.gates) {
-                    if (auto u = dynamic_cast<Ugate*>(gate)) {
+                    if (auto u = dynamic_cast<SingleQubitGate*>(gate)) {
                         std::unique_ptr<Expr> theta(RewriteExpr(u->theta, paramMap));
                         std::unique_ptr<Expr> phi(RewriteExpr(u->phi, paramMap));
                         std::unique_ptr<Expr> lambda(RewriteExpr(u->lambda, paramMap));
 
                         if (argMap.at(u->target).second == 1) {
-                            op.emplace_back<qc::StandardOperation>(nqubits, argMap.at(u->target).first, qc::U3, lambda->num, phi->num, theta->num);
+                            op.emplace_back<qc::StandardOperation>(nqubits, argMap.at(u->target).first, u->type,
+                                                                   lambda ? lambda->num : 0.,
+                                                                   phi ? phi->num : 0.,
+                                                                   theta ? theta->num : 0.);
                         } else {
                             // TODO: multiple targets could be useful here
                             for (dd::QubitCount j = 0; j < argMap.at(u->target).second; ++j) {
@@ -861,7 +881,7 @@ namespace qasm {
     void Parser::GateDecl() {
         check(Token::Kind::gate);
         // skip declarations of known gates
-        if (sym == Token::Kind::mcx_gray || sym == Token::Kind::mcx_recursive || sym == Token::Kind::mcx_vchain || sym == Token::Kind::mcphase || sym == Token::Kind::swap) {
+        if (sym == Token::Kind::mcx_gray || sym == Token::Kind::mcx_recursive || sym == Token::Kind::mcx_vchain || sym == Token::Kind::mcphase || sym == Token::Kind::swap || sym == Token::Kind::sxgate || sym == Token::Kind::sxdggate) {
             while (sym != Token::Kind::rbrace)
                 scan();
 
@@ -909,7 +929,13 @@ namespace qasm {
                 check(Token::Kind::rpar);
                 check(Token::Kind::identifier);
 
-                gate.gates.push_back(new Ugate(theta, phi, lambda, t.str));
+                gate.gates.push_back(new SingleQubitGate(t.str, qc::U3, lambda, phi, theta));
+                check(Token::Kind::semicolon);
+            } else if (sym == Token::Kind::sxgate || sym == Token::Kind::sxdggate) {
+                const auto gateType = sym == Token::Kind::sxgate ? qc::SX : qc::SXdag;
+                scan();
+                check(Token::Kind::identifier);
+                gate.gates.push_back(new SingleQubitGate(t.str, gateType));
                 check(Token::Kind::semicolon);
             } else if (sym == Token::Kind::cxgate) {
                 scan();
@@ -1024,8 +1050,8 @@ namespace qasm {
                             paramMap[gateIt->second.parameterNames[i]] = parameters[i];
 
                         for (auto& it: gateIt->second.gates) {
-                            if (auto u = dynamic_cast<Ugate*>(it)) {
-                                gate.gates.push_back(new Ugate(RewriteExpr(u->theta, paramMap), RewriteExpr(u->phi, paramMap), RewriteExpr(u->lambda, paramMap), argMap.at(u->target)));
+                            if (auto u = dynamic_cast<SingleQubitGate*>(it)) {
+                                gate.gates.push_back(new SingleQubitGate(argMap.at(u->target), u->type, RewriteExpr(u->lambda, paramMap), RewriteExpr(u->phi, paramMap), RewriteExpr(u->theta, paramMap)));
                             } else if (auto cx = dynamic_cast<CXgate*>(it)) {
                                 gate.gates.push_back(new CXgate(argMap.at(cx->control), argMap.at(cx->target)));
                             } else if (auto cu = dynamic_cast<CUgate*>(it)) {
@@ -1079,7 +1105,7 @@ namespace qasm {
                             std::vector<std::string> controls{};
                             for (size_t i = 0; i < arguments.size() - 1; ++i)
                                 controls.emplace_back(arguments[i]);
-                            if (auto u = dynamic_cast<Ugate*>(cGateIt->second.gates.at(0))) {
+                            if (auto u = dynamic_cast<SingleQubitGate*>(cGateIt->second.gates.at(0))) {
                                 gate.gates.push_back(new CUgate(RewriteExpr(u->theta, paramMap), RewriteExpr(u->phi, paramMap), RewriteExpr(u->lambda, paramMap), controls, arguments.back()));
                             } else {
                                 throw QASMParserException("Could not cast to UGate in gate declaration.");
@@ -1112,6 +1138,7 @@ namespace qasm {
     std::unique_ptr<qc::Operation> Parser::Qop() {
         if (sym == Token::Kind::ugate || sym == Token::Kind::cxgate ||
             sym == Token::Kind::swap || sym == Token::Kind::identifier ||
+            sym == Token::Kind::sxgate || sym == Token::Kind::sxdggate ||
             sym == Token::Kind::mcx_gray || sym == Token::Kind::mcx_recursive || sym == Token::Kind::mcx_vchain || sym == Token::Kind::mcphase)
             return Gate();
         else if (sym == Token::Kind::measure) {
